// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//
// Giuseppe Vacanti, Summer 1999
// European Space Agency
//
// + Scan a list of directories for CCF constituents and build a
//   suitable CCF Index File. 
// + As above, but create the CCF Master Index File.
//
#include <string> // C++
#include <vector>
#include <algorithm>
#include <cstdlib>

#include <stringutil.h> // SAS
#include <Tracer.h>
#include <param.h>
#include <STime.h>
#include <errstr.h>
#include <Directory.h>

#include <Dal.h>
#include <OalServer.h>

#include "CcfConstituent.h"
#include "CcfConstituentFactory.h"
#include "CalIndex.h"
#include "CalIndexSet.h"

#include "cifbuild.h" // generated automatically

void cifbuild() {
  PROC

  typedef std::vector<CcfConstituent *> LCC;

  string calindexset(stringParameter("calindexset"));

  // The MIF is never allowed to contain the full file path.
  bool masterindex = booleanParameter("masterindex");
  bool fullpath = booleanParameter("fullpath");
  if(masterindex){
    fullpath = false;
    Sas::message(Msg::App, Msg::Sparse, "When creating a Master Index File the value of fullpath is set to false.");
  }
  

  // These two are used if we're creating a master index file ...
  STime analysisDate("now");
  STime observationDate("1990-01-01T00:00:00");

  // ... otherwise we get them from the command line.
  if(!masterindex){
    analysisDate = STime((stringParameter("analysisdate")));
    if(booleanParameter("withobservationdate"))
      observationDate = STime(stringParameter("observationdate"));
    else {
      errstr << "Will ask the analysis date to the OAL."
	     << Sas::message(Msg::App, Msg::Silent);
      const Odf *odf = oalServer.odf();
      errstr << "Using the ODF " << setw(4) << setfill('0') << odf->orbitNr() << "_"
	     << odf->proposalId() << odf->observationId()
	     << " found in " << odf->directoryName()
	     << Sas::message(Msg::App, Msg::Silent);
      observationDate = odf->observationStartTime();

      if(booleanParameter("usecanonicalname")){
	stringstream ss;
	ss << setw(4) << setfill('0') << odf->orbitNr() << "_"
	   << odf->proposalId() << odf->observationId() 
	   << ".cif";
	calindexset = ss.str();
	// Make the name of the calindexset available to metatasks
	setParameter("calindexset", calindexset);
      }

    }

    errstr << "Observation date: " << observationDate
	   << Sas::message(Msg::App, Msg::Silent);
    errstr << "Analysis date: " << analysisDate
	   << Sas::message(Msg::App, Msg::Silent);

    if(analysisDate < observationDate)
      errstr << "Analysis date (" << analysisDate << ") < observation date ("
	     << observationDate << ") may lead to an inconsistent calibration. Proceed at you own risk."
	     << Sas::warning("PossibleInconsistentCalibration");
  } // !masterindex

  LCC allccf;    
  CalIndex * cif;

  string category = booleanParameter("ignorecategory") ?
    "any" : stringParameter("category");
    
  // The calibration index file is created from the list of CCF
  // constituents found in the directories listed on the command line
  // or given in the environment variable.
  string ccfpath;
  if(!booleanParameter("withmasterindexset")){
    
    if(booleanParameter("withccfpath"))
      ccfpath = stringParameter("ccfpath");
    else
      if(std::getenv("SAS_CCFPATH") == 0)
	errstr << "SAS_CCFPATH is not defined and I was not given an explicit ccfpath."
	       << Sas::error("NoCcfPath");
      else
	ccfpath = std::getenv("SAS_CCFPATH");
    
    vector<string> ccf_directory_list = breakup(ccfpath, ":");

    Sas::message(Msg::App, Msg::Verbose, "ccfpath: " + ccfpath);
    
    // Find directory entries that match the fileglob ...
    for(vector<string>::const_iterator i = ccf_directory_list.begin();
	i != ccf_directory_list.end(); i++) {
      const string &s = *i;
      Directory d(s, stringParameter("fileglob"), booleanParameter("recurse"));

      errstr << "Scanning " + s + " for CCF constituents." <<
	Sas::message(Msg::App, Msg::Verbose);
      
      // ... these are candidate CcfConstituents
      for(Directory::iterator j = d.begin(); j != d.end(); j++) {
	CcfConstituent * c = ccfConstituentFactory->ccfConstituent((*j)->name());
	errstr << "Found " + c->name() << Sas::message(Msg::App, Msg::Verbose);
	allccf.push_back(c);
      }	
    } // loop over ccfpath
    cif = new CalIndex(calindexset, booleanParameter("masterindex"));
  } else { // generate calibration index file from another calibration
           // index file
    CalIndexSet mif(stringParameter("masterindexset"), true);
    cif = new CalIndex(calindexset, false);
    for(unsigned i = 0; i < mif.size(); ++i)
      allccf.push_back(mif.getConstituent(i));
    mif.close();
  }

  // Now filter out CCF constituents that do not match the search
  // criteria. 
  LCC lccf;    
  for(LCC::const_iterator k = allccf.begin();
      k != allccf.end(); ++k){
    CcfConstituent *c = *k;
    if(booleanParameter("ignorecategory") ||
       c->category() == stringParameter("category")){
      // If creating a master index set, all CCF constituents are
      // pushed on the list ...
      if(masterindex)
	lccf.push_back(c);
      // ... otherwise acceptable constituents must conform to the
      // date selection criteria specified in the CCF ICD.

//        else if(c->valdate() <= observationDate && c->subdate() <= analysisDate)

      else if(c->valdate() <= observationDate &&
	      c->subdate() <= analysisDate &&
	      (c->hasEndvaldate() ? c->endvaldate() >= observationDate : true))
	lccf.push_back(c);
      else
	errstr << "This CCF constituent:\n\n" << *c
	       << "\n   does not match the search criteria:"
	       << "\n   Analysis date    = " << analysisDate
	       << "\n   Observation date = " << observationDate
	       << Sas::message(Msg::App, Msg::Verbose);
    } else 
      errstr << "This CCF constituent:\n\n" << *c
	     << "\n   does not match the search criterion:"
	     << "\n   Category         = " << category
	     << Sas::message(Msg::App, Msg::Verbose);	
  }

  // In append mode, first extracts constituents from the
  // preexistig CIF. The latter will be overwritten. Hence this
  // will not work if the DAL is not allowed to clobber an
  // existing dataset.
  if(booleanParameter("append")){
    CalIndexSet ocif(calindexset);
    for(unsigned long i=0; i<ocif.size(); i++)
      cif->addConstituent(ocif.getConstituent(i));
    ocif.close();
  }

  if(errHandler.verbosity(Msg::App) >= Msg::Noisy){
    Sas::message(Msg::App, Msg::Noisy, "List of constituents before sort:");
    for(LCC::const_iterator i = lccf.begin(); i != lccf.end(); i++)
      Sas::message(Msg::App,Msg::Noisy,(*i)->name());
  }

  // This is the sorting algorithm described in the CCF ICD	
  // (see also the definition of CcfConstituent::operator<)
  // The list is sorted in ascending order, and we scan it in
  // reverse order because we want to select the most recent
  // constituent for each scope/caltype combination.
  
  std::sort(lccf.begin(), lccf.end(), CcfConstituent::compare);

  if(errHandler.verbosity(Msg::App) >= Msg::Noisy){
    Sas::message(Msg::App, Msg::Noisy, "List of constituents after sort:");
    for(LCC::const_iterator i = lccf.begin();
	i != lccf.end(); i++)
      Sas::message(Msg::App,Msg::Noisy,(*i)->name());
  }

  LCC::iterator p = std::unique(lccf.begin(), lccf.end(),
					      CcfConstituent::equal);
  lccf.erase(p, lccf.end());

  if(errHandler.verbosity(Msg::App) >= Msg::Noisy){
    Sas::message(Msg::App, Msg::Noisy, "List of constituents after unique:");
    for(LCC::const_iterator i = lccf.begin();
	i != lccf.end(); i++)
      Sas::message(Msg::App,Msg::Noisy,(*i)->name());
  }

  for(LCC::reverse_iterator i = lccf.rbegin(); i
	!= lccf.rend(); i++) 
    cif->addConstituent(*i);
 
  cif->sort(CcfConstituent::compare); 
  if(cif->size() == 0)
    errstr <<
      "Could not find any CCF constituents matching the search criteria:\n" <<
      "\n   Analysis date    = " << analysisDate <<
      "\n   Observation date = " << observationDate <<
      "\n   ccfpath          = " << 
      "\n   Category         = " << category <<
      "\n   File glob        = " << stringParameter("fileglob") <<
      "\n   CCF search path  = " << ccfpath <<
      Sas::warning("NoMatchingCcfConstituents");
  cif->save(fullpath, analysisDate, observationDate);
  errstr << "Cif contents: \n" << *cif << Sas::message(Msg::App, Msg::Noisy);
  delete cif;  
}
